package ShowXPath.XMLParser;

import java.io.*;
import java.net.URL;
import java.net.MalformedURLException;
import java.io.*;
import java.util.*;
import java.util.Hashtable;
import java.net.URL;
import java.net.MalformedURLException;

import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.xml.sax.*;
import org.xml.sax.ErrorHandler;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.misc.*;

import ShowXPath.TreeNode.*;
import ShowXPath.StackNode.*;

import org.apache.xerces.parsers.SAXParser;

public class SaxTree
	extends HandlerBase
	implements LexicalHandler, NamespaceHandler, DTDHandler, EntityResolver, DeclHandler
{
	private Locator _locator = null;
	private Hashtable _namespace = new Hashtable();
	private LRoot _root = null;
	private LTreeNode _currentElement;
	private boolean _DTDStarted = false;

  static public void main(String[] args) {
		if (args.length != 1) {
			System.err.println("usage: java SaxTree file");
			System.exit(1);
		}
		File file =  new File(args[0]);
		if (file.isAbsolute()) {
			System.err.println(args[0] + " is not a relative file");
			System.exit(1);
		}

		SaxTree tester = new SaxTree(args[0]);
	}

	public SaxTree()
	{
		//
	}

	private void print(String string)
	{
		//System.out.println(string);
	}

	public SaxTree(URL inputFile)
	{
		Parser parser = new SAXParser();
		try
		{
			parser.setDocumentHandler(this);
			parser.setErrorHandler(this);
			parser.setDTDHandler(this);
			parser.setEntityResolver(this);
			try
			{
				((SAXParser)parser).setFeature("http://xml.org/sax/features/namespaces", true);
				((SAXParser)parser).setFeature("http://apache.org/xml/features/validation/dynamic", true);
			}
			catch (SAXException e) {
				print("error in setting up parser feature");
			} 

			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/LexicalHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/DeclHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/NamespaceHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/properties/namespace-sep", "|");
			parser.parse(inputFile.toString());
		}
		catch (SAXParseException e) {
		}
		catch (SAXException e) {
			System.err.println(inputFile + ": " + e.getMessage());
		}
		catch (IOException e) {
			System.err.println(inputFile + ": " + e.getMessage());
		}
	}

	public SaxTree(String inputFile)
	{
		File file =  new File(inputFile);
		if (file.isAbsolute()) {
			// is an absolute file (so what?)
		}

		Parser parser = new SAXParser();
		try
		{
			parser.setDocumentHandler(this);
			parser.setErrorHandler(this);
			parser.setDTDHandler(this);
			parser.setEntityResolver(this);
			try
			{
				((SAXParser)parser).setFeature("http://xml.org/sax/features/namespaces", true);
				((SAXParser)parser).setFeature("http://apache.org/xml/features/validation/dynamic", true);
			}
			catch (SAXException e) {
				print("error in setting up parser feature");
			} 

			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/LexicalHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/DeclHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/handlers/NamespaceHandler",
												 this);
			((SAXParser)parser).setProperty("http://xml.org/sax/properties/namespace-sep", "|");
			parser.parse(fileToURL(file).toString());
		}
		catch (SAXParseException e) {
		}
		catch (SAXException e) {
			System.err.println(inputFile + ": " + e.getMessage());
		}
		catch (IOException e) {
			System.err.println(inputFile + ": " + e.getMessage());
		}
	}

	public LRoot getRoot()
	{
		return _root;
	}

  static public URL fileToURL(File file) {
    String path = file.getAbsolutePath();
    String fSep = System.getProperty("file.separator");
    if (fSep != null && fSep.length() == 1)
      path = path.replace(fSep.charAt(0), '/');
    if (path.length() > 0 && path.charAt(0) != '/')
      path = '/' + path;
    try {
      return new URL("file", null, path);
    }
    catch (java.net.MalformedURLException e) {
      /* According to the spec this could only happen if the file
				 protocol were not recognized. */
      throw new Error("unexpected MalformedURLException");
    }
  }

	private LTreeNode addElement(String name)
	{
		LElement e = new LElement(name, _currentElement, _namespace);
		
		if (_currentElement instanceof LRoot)
		{
			((LRoot)_currentElement).addChild(e);
		}
		else if (_currentElement instanceof LElement)
		{
			((LElement)_currentElement).add(e);
		}
		_currentElement = e;
		return e;
	}
	
	private LTreeNode addText(String text)
	{
		print("Inside addText");
		LText e = new LText(text, _currentElement);
		e.setURI(getURI());

		if (_currentElement instanceof LElement)
		{
			print("Current element is an LElement");
			LNodeset children = ((LElement)_currentElement).getChildren();
			print("Got the children");
			if (children.size() > 0)
			{
				LTreeNode node = (LTreeNode)((LinkedList)children).getLast();
				print("Testing node now");
				if (node != null)
				{
					print("node is not null");
					if (node instanceof LText)
					{
						print("node is a LText");
						((LText)node).appendText(text);
					}
					else
					{
						print("node is not a Ltext");
						((LElement)_currentElement).add(e);
					}
				}
				else
				{
					print("node is null");
					((LElement)_currentElement).add(e);
				}
			}
			else
			{
				print("CHildren has no node in it");
				((LElement)_currentElement).add(e);
			}
		}
		return e;
	}
	
	private LTreeNode addComment(String text)
	{
		LComment e = new LComment(text, _currentElement);
		e.setURI(getURI());
		
		if (_currentElement instanceof LRoot)
		{
			((LRoot)_currentElement).addChild(e);
		}
		else if (_currentElement instanceof LElement)
		{
			((LElement)_currentElement).add(e);
		}
		return e;
	}
	
	private LTreeNode addProcessingInstruction(String target, String data)
	{
		LProcessingInstruction e = new LProcessingInstruction(target,
																													data,
																													_currentElement);
		e.setURI(getURI());
		
		if (_currentElement instanceof LRoot)
		{
			((LRoot)_currentElement).addChild(e);
		}
		else if (_currentElement instanceof LElement)
		{
			((LElement)_currentElement).add(e);
		}
		return e;
	}

	private String getURI()
	{
		String URI;
		if (_locator != null)
		{
			print("  Event URI: " + _locator.getSystemId());
			URI = _locator.getSystemId();
		}
		else
		{
			URI = "";
		}
		return URI;
	}
	
	
  public void startElement(String name, AttributeList atts)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		
		print("Element: " + name);
		LElement element = (LElement) addElement(name);
		element.setURI(getURI());
		

		String prefix;
		if (name.indexOf(':') != -1)
		{
			prefix = name.substring(0, name.indexOf(':'));
			print("  Prefix: " + prefix);
			print("  Namespace: " + (String) _namespace.get(prefix));
			element.setNodeNamespace((String)_namespace.get(prefix));
		}
		else 
		{
			element.setNodeNamespace((String)_namespace.get(""));
			print("  No prefix");
		}
		
		int len = atts.getLength();
		if (len > 0) {
			int[] v = new int[len];
			for (int i = 0; i < len; i++)
				v[i] = i;
			/* Do an insertion sort. */
			for (int i = 1; i < len; i++) {
				int n = v[i];
				String s = atts.getName(n);
				int j;
				for (j = i - 1; j >= 0; j--) {
					if (s.compareTo(atts.getName(v[j])) >= 0)
						break;
					v[j + 1] = v[j];
				}
				v[j + 1] = n;
			}
			for (int i = 0; i < len; i++)
			{
				if (!atts.getName(v[i]).startsWith("xmlns:") &&
						!atts.getName(v[i]).equals("xmlns"))
				{
					LAttribute attribute = new LAttribute(atts.getName(v[i]),
																								atts.getValue(v[i]),
																								element,
																								_namespace);
					attribute.setURI(getURI());
					element.addAttribute(attribute);
					if (_root == null)
					{
						print("  Root is null in startelement");
					}
					else
					{
						print("  Root is not null in startelement");
					}
					
					if (_root.isIDAttribute(name, atts.getName(v[i])))
					{
						_root.setElementID(atts.getValue(v[i]), element);
						element.setXMLID(atts.getValue(v[i]));
					}
					print("    Attr[" + i + "] = " + atts.getName(v[i]));
					print("    Value: " + atts.getValue(v[i]));
				}
				if (atts.getName(v[i]).equals("xml:lang"))
				{
					element.setLang(atts.getValue(v[i]));
				}
			}
			for (int i = 0; i < len; i++)
			{
				if (atts.getName(v[i]).startsWith("xmlns:") ||
						atts.getName(v[i]).equals("xmlns"))
				{
					LNamespace namespace = new LNamespace(atts.getName(v[i]),
																								atts.getValue(v[i]),
																								element);
					namespace.setURI(getURI());
					element.addNamespace(namespace);
					print("    Namespace[" + i + "] = " + atts.getName(v[i]));
					print("    URI: " + atts.getValue(v[i]));
				}
			}
		}
	}

  public void ignorableWhitespace(char[] cbuf, int start, int len)
	{
		if (_DTDStarted)
		{
			return;
		}
    characters(cbuf, start, len);
		print("ignoreWhiteSpace: " + String.copyValueOf(cbuf, start, len));
  }

  public void characters(char[] cbuf, int start, int len)
	{
		if (_DTDStarted)
		{
			return;
		}
		addText(String.copyValueOf(cbuf, start, len));
		print("Text Node: " + String.copyValueOf(cbuf, start, len));
  }

  public void endElement(String name)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		_currentElement = _currentElement.getParent();
		print("End Element: " + name);
  }

  public void processingInstruction(String target, String data)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		addProcessingInstruction(target, data);
		print("Processing Instruction: <?" + target + " " + data + "?>");
  }

  public void startDocument()
	{
		if (_DTDStarted)
		{
			return;
		}
		print("Start Document: ");
		if (_locator != null)
		{
			print("  Event in : " + _locator.getSystemId());
		}
		_root = new LRoot();
		_root.setURI(_locator.getSystemId());
		_currentElement = _root;
	}

  public void endDocument() throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		print("End Document");
  }

	public void startDTD(String name, String publicId, String systemId)
		throws SAXException
	{
		print("Start DTD: " + name + " PUBLIC " + publicId + " SYSTEM "
					+ systemId);
		_root.startDTD(name, publicId, systemId);
		_DTDStarted = true;
	}
	
	public void endDTD ()
		throws SAXException
	{
		print("End DTD");
		_DTDStarted = false;
	}

	public void startEntity (String name)
		throws SAXException
	{
		print("Start Entity:" + name);
	}

	public void endEntity (String name)
		throws SAXException
	{
		print("End Entity: " + name);
	}
	public void startCDATA ()
		throws SAXException
	{
		print("Start CDATA");
	}
	
	public void endCDATA ()
		throws SAXException
	{
		print("End CDATA");
	}

	public void comment (char ch[], int start, int length)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		addComment(String.copyValueOf(ch, start, length));
		print("Comment:" + String.copyValueOf(ch, start, length));
	}

	public void elementDecl (String name, String model)
		throws SAXException
	{
		print("Element Declaration: " + name + " model: " + model);
		_root.elementDecl(name, model);
	}

	public void attributeDecl (String eName,
														 String aName,
														 String type,
														 String valueDefault,
														 String value)
		throws SAXException
	{
		print("Attrlist Element Name: " + eName);
		print("  Attribute Name: " + aName);
		print("  Type: " + type);
		print("  Default Value: " + valueDefault);
		print("  Value: " + value);
		_root.attributeDecl(eName, aName, type, valueDefault, value);
	}

	public void internalEntityDecl (String name, String value)
		throws SAXException
	{
		print("Internal Entity Declaration: " + name);
		print("  Value: " + value);
	}

	public void externalEntityDecl (String name, String publicId,
																	String systemId)
		throws SAXException
	{
		print("External Entity Declaration: " + name);
		print("  PUBLIC: " + publicId);
		print("  SYSTEM: " + systemId);
	}

	public void startNamespaceDeclScope (String prefix, String uri)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		print("Start Namespace: " + prefix);
		print("  URI: " + uri);
		_namespace.put(prefix, uri);
	}

	public void endNamespaceDeclScope (String prefix)
		throws SAXException
	{
		if (_DTDStarted)
		{
			return;
		}
		print("End Namespace: " + prefix);
		_namespace.remove(prefix);
	}

  public void notationDecl (String name,
														String publicId,
														String systemId)
	{
		print("Notation Declaration: " + name);
		print("  PUBLIC: " + publicId);
		print("  SYSTEM: " + systemId);
	}

  public void unparsedEntityDecl (String name,
																	String publicId,
																	String systemId,
																	String notationName)
	{
		_root.addUnparsedEntity(name, systemId, publicId, notationName);
		print("Unparsed Entity Declaration: " + name);
		print("  PUBLIC: " + publicId);
		print("  SYSTEM: " + systemId);
		print("  NOTATION: " + notationName);
	}

  public InputSource resolveEntity (String publicId,
																		String systemId)
	{
		print("Resolve Entity");
		print("  PUBLIC: " + publicId);
		print("  SYSTEM: " + systemId);
		return new InputSource(systemId);
	}

	public void setDocumentLocator(Locator locator)
	{
		_locator = locator;
	}
	
}
